/*
  Midnight, a C++ midi library to load and play midi files.
  Copyright (C) 1999 Rodrigo Reyes, reyes@chez.com

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <iostream.h>

#include "IOUtils.h"
#include <stdlib.h>
#include <string.h>

#ifndef _MIDIEVENT_H_
#define _MIDIEVENT_H_

/// A basic, abstract, and rather unuseful class for the data of midi events
class MidiAbstractDataEvent { };

/// This is the standard MidiEvent. Typically, a midi track should be full of them.
class MidiEvent
{
 public:

  ///
  enum {
    UNKNOWN, 

    NOTE_OFF, NOTE_ON, KEY_AFTER_TOUCH, CONTROL_CHANGE, PROG_CHANGE,
    CHANNEL_AFTER_TOUCH, PITCH_WHEEL_CHANGE,

    META_UNKNOWN, 
    META_SEQUENCE_NUMBER, META_TEXT_EVENT, META_COPYRIGHT, 
    META_SEQUENCE_NAME, META_TRACK_INSTRUMENT_NAME, META_LYRIC,
    META_MARKER, META_CUE_POINT, META_END, META_SET_TEMPO,
    META_TIME_SIG, META_KEY_SIG, META_SPECIFIC,

    SYSEX

  } EVENTCODE;

 private:
  /** The deltatime is the number of tick to wait before the event. 
      This is a delta, so it is only the difference between the
      previous event, and the current one.  
  */
  int _deltaTime;
  /** This is the total number of ticks to wait, from the start of the music.
   */
  unsigned long _offset;
  /// The type of the event. This is a value from the EVENTCODE enum.
  unsigned int _eventCode;

  /// The data associated to the event. This is dependent from the _eventCode variable.
  void *_data;

  /// Initializes the data of the object.
  void init(istream &in, int &length, unsigned long offset, int &runningStatus);

 public:

  /** Creates a new MidiEvent.
      If this is the first event of a track, the offset shall be 0, or the offset the previous event otherwise.
      @param in the istream to build the event from.
      @param offset the offset of the previous event.
  */
  MidiEvent(istream &in, unsigned long offset)
    {
      int dummy,dummy2=-1;
      init(in, dummy,offset, dummy2);
    }
  
  /** Creates a new MidiEvent.
      If this is the first event of a track, the offset shall be 0, or the offset the previous event otherwise.
      @param in the istream to build the event from.
      @param length this variable is incremented with the number of bytes read in the stream to build the event.
      @param offset the offset of the previous event.
      @param runningStatus the midi running status, set up by the constructor for the next event, or by the previous event for the current one.
  */
  MidiEvent(istream &in, int &length, unsigned long offset, int &runningStatus)
    {
      init(in,length,offset, runningStatus);
    }

  ///
  ~MidiEvent()
    {
      delete _data;
    }

  /// Returns the delta of the event.
  int getDelta() { return _deltaTime; }
  /// Returns the event code as defined by the EVENTCODE enum.
  int getEventCode() { return _eventCode; }
  /// Returns the offset of the event in the track.
  unsigned long getOffset() { return _offset; }
  /// Returns the data structure of the midi event.
  void *getData() { return _data; }
};


///
class MidiEventCommand : public MidiAbstractDataEvent
{
 private:
  ///
  char _channel;
  ///
  char _data1, _data2;

 public:
  ///
  MidiEventCommand(char channel, char data1, char data2)
    {
      _channel = channel;
      _data1 = data1;
      _data2 = data2;
    }

  ///
  char getChannel() { return _channel; }
  ///
  char getData1() { return _data1; }
  ///
  char getData2() { return _data2; }

}///
;


///
class MidiMetaEvent : public MidiAbstractDataEvent
{
  ///
  char _command;
  ///
  char _data1, _data2;
 public:
  ///
  MidiMetaEvent(char command, char data1, char data2)
    {
      _command = command;
      _data1 = data1;
      _data2 = data2;
    }
  ///
  int getCommand() { return _command; }
  ///
  int getData1() { return _data1; }
  ///
  int getData2() { return _data2; }
}///
;


///
class MidiMetaSequenceNumber : public MidiAbstractDataEvent
{
  ///
  int _length;
  ///
  int *_sequence;
 public:
  ///
  MidiMetaSequenceNumber(int length)
    {
      _length = length;
      _sequence = new int[length];
    }
  ///
  ~MidiMetaSequenceNumber()
    {
      delete[] _sequence;
    }
  ///
  int getLength() { return _length; }
  ///
  void setSequence(int idx, int value) { _sequence[idx] = value; }
  ///
  int getSequence(int idx) { return _sequence[idx]; }
  ///
  int* getSequence() { return _sequence; }
}///
;


///
class MidiMetaTempo : public MidiAbstractDataEvent
{
  ///
  int _tempo;
 public:
  ///
  MidiMetaTempo(int tempo) { _tempo = tempo; }
  ///
  int getTempo() { return _tempo; }
}///
;


///
class MidiMetaTimeSignature : public MidiAbstractDataEvent
{
  ///
  int _numerator;
  ///
  int _denominator;
  ///
  int _clicks;
  ///
  int _quarterNotes;
 public:
  ///
  MidiMetaTimeSignature(int numerator, int denominator, int clicks, int quarters)
    {
      _numerator = numerator;
      _denominator = denominator;
      _clicks = clicks;
      _quarterNotes = quarters;
    }
  ///
  int getNumerator() { return _numerator; }
  ///
  int denominator() { return _denominator; }
  ///
  int getClicks() { return _clicks; }
  ///
  int getQuarterNotes() { return _quarterNotes; }
}///
;


///
class MidiMetaTextEvent : public MidiAbstractDataEvent
{
  ///
  int _length;
  ///
  char* _text;

 public:
  ///
  MidiMetaTextEvent(int len, const char *text)
    {
      _length = len;
      _text = IOUtils::cppstrdup(text); 
    }

  ///
  MidiMetaTextEvent(int len)
    {
      _length = len;
      _text = new char[len+1];
    }
  
  ///
  ~MidiMetaTextEvent()
    {
      delete[] _text;
    }

  ///
  int getLength() { return _length; }
  ///
  const char* getText() { return _text; }
  ///
  char* getDataBuffer() { return _text; }
}///
;


///
class MidiMetaKeySignature : public MidiAbstractDataEvent
{
  ///
  char _sharp;
  ///
  char _major;
 public:
  ///
  MidiMetaKeySignature(char sharp, char major)
    {
      _sharp = sharp;
      _major = major;
    }
  ///
  int getSharp() { return _sharp; }
  ///
  int getMajor() { return _major; }
}///
;


///
class MidiGenericEvent  : public MidiAbstractDataEvent
{
  ///
  int _code;
  ///
  int _length;
  ///
  char *_data;
 public:
  ///
  MidiGenericEvent(int code, int length)
    {
      _code = code;
      _length = length;
      _data = new char[length+1];
    }
  ///
  ~MidiGenericEvent()
    {
      delete[] _data;
    }
  ///
  char* getBuffer() { return _data; }
  ///
  int getLength() { return _length; }
  ///
  int getCode() { return _code; }
}///
;


///
ostream &operator << (ostream &o, MidiEvent *me);


#endif

