/*
  Midnight, a C++ midi library to load and play midi files.
  Copyright (C) 1999 Rodrigo Reyes, reyes@chez.com

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "IOUtils.h"
#include "MidiEvent.h"
#include <iostream.h>
#include <vector>

#ifndef __MIDIFILE_H_
#define __MIDIFILE_H_

#define SINGLE_MULTI_CHANNEL 0;
#define SYNCHRONOUS_MULTI_CHANNEL 1
#define ASYNCHRONOUS_MULTI_CHANNEL 2
#define INVALID_MIDI_FILE -1;

class MidiTrackEventIterator;


  /** Represents a single midi track.
 
    A MidiTrack object, just like a typical midi track, represents a
    sequence of midi events.  
  */

class MidiTrack
{
  friend class MidiTrackEventIterator;
  vector<MidiEvent*> _events;
public:
  MidiTrack(istream &in);
  ~MidiTrack() 
    {
      for (unsigned int i=0; i<_events.size(); i++)
	delete _events[i];
    }
};


/** Represents a midi file,  a set of midi tracks with additional informations.
 */

class MidiFile
{
  /** the midi filetype gives the topology and nature of the midi file.
    0, single track : this midi file contains just one track.
    1, multiple synchronous tracks that must be played concurrently.
    2, multiple asynchronous tracks. This is currently not supported by the midnight library.
   */
  int _midiFileType;

  /// The number of tracks for this file.
  int _nbTracks;

  /** Quoting Dave Oppenheim's version of the specifications : "the
    division of a quarter-note represented by the delta-times in the
    file.  (If division is negative, it represents the division of a
    second represented by the delta-times in the file, so that the
    track can represent events occurring in actual time instead of
    metrical time.  It is represented in the following way: the upper
    byte is one of the four values -24, -25, -29, or -30,
    corresponding to the four standard SMPTE and MIDI time code
    formats, and represents the number of frames per second."  */
  int _division;

  /** The array of tracks. 
      The length is given by the _nbTracks variable.
  */
  MidiTrack** _tracks;

public:
  /// Constructs an instance from a valid istream object.
  MidiFile(istream &in);
  /// Clean up.
  ~MidiFile() 
    {
      for (int i=0; i<_nbTracks; i++)
	delete _tracks[i];
      delete[] _tracks; 
    }

  /// Returns the midi filetype.
  int getType();

  /// Returns the number of available tracks in this file.
  int getTrackCount() { return _nbTracks; }

  /** Get a given MidiTrack in the file.
      @param idx the index of the MidiTrack in the array. No check is done on this parameter, so it will coredump if the value is wrong.
  */
  MidiTrack *getTrack(int idx) { return _tracks[idx]; }

  /** Get the division of a quarter note for this file. 
   @see the _division object variable for more documentation
  */
  int getDivision() { return _division; }

  /// Return true if the file is a valid midi file representation, false otherwise.
  bool isValid() { return (_midiFileType>=0) && (_midiFileType<2); }
};

///
class MidiTrackEventIterator
{
  MidiTrack *_track;
  MidiFile  *_file;
  vector<MidiEvent*>::iterator _iterator;

 public:
  ///
  MidiTrackEventIterator(MidiFile *mf, MidiTrack *track) : _track(track), _file(mf)
    {
      _iterator = track->_events.begin();
    }
  ///
  bool hasNext()
    {
      if (_iterator == _track->_events.end())
	return false;
      return true;
    }
  ///
  MidiEvent *getNext()
    {
      return *_iterator++;
    }
  ///
  MidiEvent *tasteNext()
    {
      return *_iterator;
    }
  ///
  const MidiFile *getMidiFile() { return _file; }

};

#endif
