/*
  Midnight, a C++ midi library to load and play midi files.
  Copyright (C) 1999 Rodrigo Reyes, reyes@chez.com

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "MidiFile.h"
#include "MidiEvent.h"
#include "MidiDevice.h"
#include "MidiUtils.h"

#ifndef __MIDIPLAYER_H_
#define __MIDIPLAYER_H_


#ifdef DEBUG_TIMING
#define DBUG_TIMING(x) x
#define DBUG_TIMING2(x,y) x=y
#else
#define DBUG_TIMING(x)
#define DBUG_TIMING2(x,y)
#endif

/// 
class MidiPlayer
{
 private:

  ///
  MidiFile*   _midiFile;  /// The midi file object
  MidiDevice* _device;  /// The output device 
  vector<pair<unsigned long, MidiEvent*> > _events; /// the vector of events

  bool _playing;  /// a flag to stop the playing
  bool _isPossiblyPlaying; // true if the player may still be playing,
 /// false if it is 100% sure it is not playing anything.

  bool _started; /// if true, play() shall initialize the variables

 protected:
  vector<pair<unsigned long, MidiEvent*> >::iterator _curEvent; /// Points to the current event

  void init()
    {
      _eventNoteOn = false;
      _eventNoteOff = false;
      _eventKeyAfterTouch = false;
      _eventControlChange = false;
      _eventProgChange = false;
      _eventChannelAfterTouch = false;
      _eventPitchWheel = false;
      _eventTempo = false;
      _eventLyric = false;
      _eventText = false;
      _eventTrackInstrumentName = false;
      _eventAny = false;
      _eventEnd = false;

      _isPossiblyPlaying = false;
      _playing = false;
      _started = false;
      

      // DEBUG the events
      //      _eventNoteOn = true;
      //      _eventNoteOff = true ;
      //      _eventKeyAfterTouch = true;
      //      DEBUG_TIMING( _eventPitchWheel = true);
      
      //      DEBUG_TIMING( _eventControlChange = true);
      //      DEBUG_TIMING( _eventProgChange = true);
      //      DEBUG_TIMING( _eventChannelAfterTouch = true);
    }

 public:

  ///
  MidiPlayer(MidiDevice *device) :  _device(device)
    {
      init();
      _midiFile = NULL;
    }

  ///
  MidiPlayer(MidiFile *file, MidiDevice *device) : _midiFile(file), _device(device)
    {
      MidiUtils::uniquifyTrack(file, _events);
      init();
      _curEvent = _events.begin();
    }

  ///
  virtual void setFile(MidiFile *file);

  ///
  virtual ~MidiPlayer() { }


  ///
  virtual vector<pair<unsigned long, MidiEvent*> > &getEvents()
    {
      return _events;
    }

  ///
  virtual MidiDevice*  getDevice() { return _device; }
  ///
  virtual void         setDevice(MidiDevice *dev) 
    { 
      _started = true;
      _device = dev; 
    }
  ///
  virtual MidiFile*    getFile() { return _midiFile; }
  
  ///
  virtual void stop();
  ///
  virtual void play();
  ///
  virtual void pause();

  ///
  virtual unsigned long totalTicks();
  ///
  virtual void setPosition(unsigned long ticks);

  ///
  virtual void setPlaying(bool val) {_playing = val; }
  ///
  virtual bool isPlaying() { return _playing; }
  ///
  virtual bool isPossiblyPlaying() { return _isPossiblyPlaying; }

 protected:
  ///
  virtual void setPossiblyPlaying(bool val) { _isPossiblyPlaying = val; }

  ///
  bool _eventNoteOn;
  ///
  virtual void eventNoteOn(vector<pair<unsigned long, MidiEvent*> >::iterator i)  
    { 
      cerr << "Note On : " << i->first << endl;
    }

  ///
  void playEvent(MidiEvent *event, bool activeEvents);

  ///
  vector<pair<unsigned long, MidiEvent*> >::iterator updateMetaEvents(unsigned long ticks)
    {
      vector<pair<unsigned long, MidiEvent*> >::iterator curev = _events.begin(); 

      if (curev == _events.end())
	return _events.end();

      while ((curev->first < ticks) && (curev != _events.end()))
	{
	  switch (curev->second->getEventCode())
	    {
	    case MidiEvent::CONTROL_CHANGE:
	    case MidiEvent::PROG_CHANGE:	  
	    case MidiEvent::META_SET_TEMPO:
	      playEvent(curev->second, false);
	      break;
	    }
	  ++curev;
	}
      return curev;
    }

  ///
  bool _eventNoteOff;
  ///
  virtual void eventNoteOff(vector<pair<unsigned long, MidiEvent*> >::iterator i) { cerr << "Note Off : " << i->first << endl; }

  ///
  bool _eventKeyAfterTouch;
  ///
  virtual void eventKeyAfterTouch(vector<pair<unsigned long, MidiEvent*> >::iterator i) { cerr << "After Touch : " << i->first << endl; }

  ///
  bool _eventControlChange;
  ///
  virtual void eventControlChange(vector<pair<unsigned long, MidiEvent*> >::iterator i) { }

  ///
  bool _eventProgChange;
  ///
  virtual void eventProgChange(vector<pair<unsigned long, MidiEvent*> >::iterator i) { }

  ///
  bool _eventChannelAfterTouch;
  ///
  virtual void eventChannelAfterTouch(vector<pair<unsigned long, MidiEvent*> >::iterator i) { }

  ///
  bool _eventPitchWheel;
  ///
  virtual void eventPitchWheel(vector<pair<unsigned long, MidiEvent*> >::iterator i) { }

  ///
  bool _eventTempo;
  ///
  virtual void eventTempo(vector<pair<unsigned long, MidiEvent*> >::iterator i) { }

  ///
  bool _eventLyric;
  ///
  virtual void eventLyric(vector<pair<unsigned long, MidiEvent*> >::iterator i)   { }

  ///
  bool _eventText;
  ///
  virtual void eventText(vector<pair<unsigned long, MidiEvent*> >::iterator i)    { }

  ///
  bool _eventTrackInstrumentName;
  ///
  virtual void eventTrackInstrumentName(vector<pair<unsigned long, MidiEvent*> >::iterator i)    { }

  ///
  bool _eventAny;
  ///
  virtual void eventAny(vector<pair<unsigned long, MidiEvent*> >::iterator i)    { }

  ///
  bool _eventEnd;
  ///
  virtual void eventEnd(vector<pair<unsigned long, MidiEvent*> >::iterator i)    { }

}///
;

#endif
