/*
  Midnight, a C++ midi library to load and play midi files.
  Copyright (C) 1999 Rodrigo Reyes, reyes@chez.com

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "MidnightConfiguration.h"
#include "SynthOSSDevice.h"
#include "AWEDevice.h"
#include "MidiUtils.h"
#include "ExtOSSDevice.h"

#include <fstream>
#include <strstream>
#include <stdlib.h>

char * MidnightConfiguration::_defaultDeviceName = "/dev/sequencer";

MidnightConfiguration::MidnightConfiguration(const char *filename) 
{
  _confFileName = MidiUtils::cppstrdup(filename);
  _deviceName = NULL;
  _curDevice = NULL;

  setDeviceName(_defaultDeviceName);
  setDefault();

  if (load() == false)
    {
      cerr << "Configuration file not found. Creating a default " << filename << endl;
      setDefault();
      save();
    }
}

MidnightConfiguration::~MidnightConfiguration()
{
  delete[] _confFileName;
}

bool splitLine(const char *buffer, char*name, char*value)
{
  if ((buffer[0]=='\0') || (buffer[0] == '#'))
    return false;

  while ((*buffer==' ')||(*buffer=='\t')) ++buffer;

  while ( ((*buffer>='a')&&(*buffer<='z')) || ((*buffer>='A')&&(*buffer<='Z'))
	  || (*buffer == '/') || (*buffer=='_') 
	  || ((*buffer>='0') && (*buffer<='9')))
    {
      *name++ = *buffer++;
    }
  *name = 0;

  while ((*buffer==' ')||(*buffer=='\t')) ++buffer;

  if (*buffer++ != '=')
    return false;

  while ((*buffer==' ')||(*buffer=='\t')) ++buffer;
  
  while ( *buffer > 32 )
    {
      *value++ = *buffer++;
    }
  *value = 0;

  while ((*buffer==' ')||(*buffer=='\t')) ++buffer;

  if ((*buffer=='\0')||(*buffer=='\n'))
    return true;

  return false;
}

bool MidnightConfiguration::load()
{
  ifstream fin(_confFileName);

  if (!fin)
    return false;

  {
    char buffer[1024];
    char name[256], value[1024];
  
    while(fin.getline(buffer,1024))
      {
	if ((buffer[0] != '#') && (strlen(buffer)>0))
	  {
	    splitLine(buffer, name, value);
	    _features[name] = value;
	  }
      }
  }

  //
  // setup the internal values
  //
  map<string,string>::iterator iter;

  if ((iter = _features.find("DEVICE")) != _features.end())
    {
      setDeviceName(iter->second.c_str());
    }
  if ((iter = _features.find("NUMBER")) != _features.end())
    {
      _deviceNumber = atoi(iter->second.c_str());
    }
  if ((iter = _features.find("TYPE")) != _features.end())
    {
      const char *value = iter->second.c_str();
      if (strcmp(value,"OSS_INTERNAL_SYNTH")==0)
	_deviceType = OSS_INTERNAL_SYNTH;
      else if (strcmp(value, "OSS_AWE_SYNTH")==0)
	_deviceType = OSS_AWE_SYNTH;
      else if (strcmp(value, "OSS_EXTERNAL_MIDI")==0)
	_deviceType = OSS_EXTERNAL_MIDI;
      else if (strcmp(value, "GUESS_BEST")==0)
	_deviceType = GUESS_BEST;
    }
  if ((iter = _features.find("OUTPUT")) != _features.end())
    {
      if (strcmp(iter->second.c_str(),"INT")==0)
	_output = SYNTH_OUTPUT;
      else if (strcmp(iter->second.c_str(), "EXT")==0)
	_output = MIDI_OUTPUT;
    }

  fin.close();
  return true;
}

/*
bool MidnightConfiguration::load()
{
  ifstream fin(_confFileName);

  if (!fin)
    return false;

  char buffer[256];
  char name[256], value[256];

  while (fin.getline(buffer, 256) )
    {
      if ((buffer[0] != '#') && (strlen(buffer)>0))
	{
	  splitLine(buffer, name, value);

	  if ((strcmp(name, "DEVICE") == 0))
	    {
	      setDeviceName(value);
	    }
	  else if (strcmp(name, "NUMBER") == 0)
	    {
	      _deviceNumber = atoi(value);
	    }
	  else if (strcmp(name, "TYPE") == 0)
	    {
	      if (strcmp(value,"OSS_INTERNAL_SYNTH")==0)
		_deviceType = OSS_INTERNAL_SYNTH;
	      else if (strcmp(value, "OSS_AWE_SYNTH")==0)
		_deviceType = OSS_AWE_SYNTH;
	      else if (strcmp(value, "OSS_EXTERNAL_MIDI")==0)
		_deviceType = OSS_EXTERNAL_MIDI;
	      else if (strcmp(value, "GUESS_BEST")==0)
		_deviceType = GUESS_BEST;
	    }
	  else if (strcmp(name, "OUTPUT") == 0)
	    {
	      if (strcmp(value,"INT")==0)
		_output = SYNTH_OUTPUT;
	      else if (strcmp(value, "EXT")==0)
		_output = MIDI_OUTPUT;
	    }

	}
    }
    
  fin.close();

  return true;
}
*/


bool MidnightConfiguration::save()
{
  ofstream fout(_confFileName);

  char buffer[256];
  _features["DEVICE"] = _deviceName;
  sprintf(buffer,"%d",_deviceNumber);
  _features["NUMBER"] = buffer;
  switch(_deviceType)
    {
    case OSS_INTERNAL_SYNTH:
      _features["TYPE"] = "OSS_INTERNAL_SYNTH";
      break;
    case OSS_AWE_SYNTH:
      _features["TYPE"] = "OSS_AWE_SYNTH";
      break;
    case OSS_EXTERNAL_MIDI:
      _features["TYPE"] = "OSS_EXTERNAL_MIDI";
      break;
    case GUESS_BEST:
      _features["TYPE"] = "GUESS_BEST";
      break;
    }
  if (_output == SYNTH_OUTPUT)
    {
      _features["OUTPUT"] = "INT";
    }
  else 
    {
      _features["OUTPUT"] = "EXT";
    }

  fout << "# Midnight file configuration" << endl;
  fout << "#" << endl << "# this file is generated by the midnight midi player configuration system" << endl << "#" << endl;
  fout << endl;

  for (map<string,string>::iterator iter=_features.begin(); iter != _features.end(); iter++)
    {
      string &key = iter->first;

      if (key == "DEVICE")
	{
	  fout << "# the device is usually /dev/sequencer" << endl;
	  fout << "DEVICE = " << iter->second << endl << endl;
	}
      else if (key == "NUMBER")
	{
	  fout << "# number is the device number as specified by a 'cat /dev/sndstat'" << endl;
	  fout << "# set it to -1 if you don't know" << endl;
	  fout << "NUMBER = " << iter->second << endl << endl;
	}
      else if (key == "TYPE")
	{
	  fout << "# available types are : GUESS_BEST, OSS_INTERNAL_SYNTH, OSS_AWE_SYNTH, OSS_EXTERNAL_MIDI" << endl;
	  fout << "TYPE = "  << iter->second << endl << endl;
	}
      else if (key == "OUTPUT")
	{
	  fout << "# OUTPUT = INT in order to use the internal midi synth (if available)" << endl;
	  fout << "# or OUTPUT = EXT to use an external midi output" << endl;
	  fout << "OUTPUT = " << iter->second << endl << endl;
	}
      else if (key.length()>0)
	{
	  fout << "#" << endl << key << " = " << iter->second << endl;
	}
    }

  fout.flush();
  fout.close();
  return true;
}

void MidnightConfiguration::setDefault()
{
  _output = SYNTH_OUTPUT;
  _deviceNumber = -1;
  _deviceType = GUESS_BEST;
  _output = SYNTH_OUTPUT;

  setDeviceName(_defaultDeviceName);

}

void MidnightConfiguration::setDeviceName(const char *name)
{
  delete _deviceName;
  _deviceName = MidiUtils::cppstrdup(name);
  callListeners();
}

MidiDevice* MidnightConfiguration::lockDevice()
{
  return lockDevice(_deviceType);
}

MidiDevice* MidnightConfiguration::lockDevice(int devtype)
{
  MidiDevice *dev = NULL;

  switch(devtype)
    {
    case GUESS_BEST:
      return lockDevice( guessBestType(_deviceName, _deviceNumber) );

    case OSS_INTERNAL_SYNTH:
      dev = new SynthOSSDevice(_deviceName, _deviceNumber);
      break;

    case OSS_AWE_SYNTH:
      dev = new AweDevice(_deviceName, _deviceNumber);
      break;

    case OSS_EXTERNAL_MIDI:
      dev = new ExtOSSDevice(_deviceName, _deviceNumber);
      break;
    }

  if (dev != NULL)
    internalLockDevice(dev);

  return   _curDevice = dev;
}

void MidnightConfiguration::unlockDevice(MidiDevice* dev)
{
  internalUnlockDevice(dev);
}


//
//
// dev : the device name (eg. /dev/sequencer)
// nb  : the midi device id, or -1 to pick up any
// return : the device type
int MidnightConfiguration::guessBestType(char *dev, int nb)
{
  int seq = open(dev, O_RDONLY);
  int result = UNKNOWN_TYPE;

  if (seq  == -1)
    return UNKNOWN_TYPE;

  if (_output == SYNTH_OUTPUT)
    {

      int nbSynth;
      if (ioctl(seq, SNDCTL_SEQ_NRSYNTHS, &nbSynth) == -1)
	exit(9);

      //
      // 
      if (((nb<0) || (nb < nbSynth)) && (nbSynth>0))
	{
	  synth_info sinfo;

	  if (nb >= 0)
	    sinfo.device = nb;
	  else
	    sinfo.device = nbSynth-1;

	  if (ioctl(seq, SNDCTL_SYNTH_INFO, &sinfo) != -1)
	    {
	      switch(sinfo.synth_type)
		{
		case SYNTH_TYPE_SAMPLE:
		  switch (sinfo.synth_subtype)
		    {
		    case SAMPLE_TYPE_AWE32:
		      result = OSS_AWE_SYNTH;
		      break;
		    default:
		      result = OSS_INTERNAL_SYNTH;
		    }
		  break;

		case SYNTH_TYPE_FM:
		  break;

		case SYNTH_TYPE_MIDI:
		  result =  OSS_EXTERNAL_MIDI;
		  break;

		default:

		  break;
		  
		}
	    }
	}
    }
  else 
    {
      result = OSS_EXTERNAL_MIDI;
    }
  close(seq);
      
  return result;
}

void MidnightConfiguration::setFeature(string key, string value)
{
  _features[key] = value;
}

string MidnightConfiguration::getFeature(string key, string def)
{
  map<string,string>::iterator res = _features.find(key);
  if (res != _features.end())
    return res->second;
  else
    return def;
}

void MidnightConfiguration::setIntFeature(string key, int value)
{
  char buffer[32];
  sprintf(buffer,"%d", value);
  _features[key] = buffer;
}

int MidnightConfiguration::getIntFeature(string key, int def)
{
  map<string,string>::iterator res = _features.find(key);
  if (res != _features.end())
    return atoi(res->second.c_str());
  else
    return def;  
}
