/*
  Midnight, a C++ midi library to load and play midi files.
  Copyright (C) 1999 Rodrigo Reyes, reyes@chez.com

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#ifndef _MIDNIGHTPLAYERMANAGER_H_
#define _MIDNIGHTPLAYERMANAGER_H_

#include "pthread.h"
#include "ThreadedQueue.h"
#include <fstream>
#include <gtk/gtk.h>
#include <iostream.h>
#include <sys/stat.h>
#include <stdio.h>
#include <unistd.h>
#include <vector>
#include <string>
#include "MidiEvent.h"
#include "MidiUtils.h"

#include "MidiPlayer.h"
#include "MidnightConfiguration.h"
#include "MidiFile.h"
#include "MidiDevice.h"

struct InfoBag
{
  int Command;
  int Value;
  string Buffer;

  InfoBag(int cmd, int val) : Command(cmd), Value(val)
    { 
    }

  InfoBag(int cmd, int val, string str) : Command(cmd), Value(val), Buffer(str)
    {
    }
};

ostream &operator << (ostream &o, InfoBag&bag);

class MidnightPlayerManager : public MidnightConfigurationListener
{
 private:

  ThreadedQueue<InfoBag> _playerqueue;
  ThreadedQueue<InfoBag> _controllerqueue;

  MidiPlayer*               _player;
  MidnightConfiguration*    _config;
  MidiFile*                 _file;
  MidiDevice*               _device;

  enum {
    PLAYER_LOAD, PLAYER_PLAY, PLAYER_PAUSE, PLAYER_STOP, PLAYER_SET_POSITION
  } PlayerCommand;

  enum {
    CONTROLLER_NOTHING, CONTROLLER_TICK, CONTROLLER_TICK_LYRICS, CONTROLLER_TOTAL_LENGTH, CONTROLLER_CLEAR_LYRICS, 
    CONTROLLER_ADD_LYRICS, CONTROLLER_ADD_LYRICS_END, CONTROLLER_SONG_ENDED,
    CONTROLLER_CLEAR_INFO, CONTROLLER_ADD_INFO, CONTROLLER_ADD_INFO_END,
    CONTROLLER_CLEAR_TITLE, CONTROLLER_ADD_TITLE, CONTROLLER_ADD_TITLE_END,
    CONTROLLER_SET_LANGUAGE, CONTROLLER_UPDATE_INFO
  } ControllerCommand;

  //
  // Specializes the MidiPlayer class, and makes it thread-safe (at
  // least for the derived methods)
  //
  class LocalMidiPlayer : public MidiPlayer
    {
      MidnightPlayerManager* _manager;
      pthread_mutex_t _mutex;

    public:


      LocalMidiPlayer(MidnightPlayerManager *manager, MidiFile *file, MidiDevice *device) : MidiPlayer(file,device), _manager(manager)
	{
	  _eventAny = true;
	  _eventLyric = true;
	  _eventText = true;
	  _eventEnd  = true;
	  pthread_mutex_init (&_mutex, NULL);
	}

      
      virtual bool isPlaying() 
	{
	  pthread_mutex_lock (&_mutex);
	  
	  bool result = MidiPlayer::isPlaying();
	  
	  pthread_mutex_unlock (&_mutex);
	  return result;
	}

      virtual void setPlaying(bool val)
	{
	  pthread_mutex_lock (&_mutex);
	  
	  MidiPlayer::setPlaying(val);
	  
	  pthread_mutex_unlock (&_mutex);
	}

      virtual void setPossiblyPlaying(bool val)
	{
	  pthread_mutex_lock (&_mutex);
	  
	  MidiPlayer::setPossiblyPlaying(val);
	  
	  pthread_mutex_unlock (&_mutex);
	}

      virtual bool isPossiblyPlaying() 
	{ 
	  pthread_mutex_lock (&_mutex);
	  
	  bool result = MidiPlayer::isPossiblyPlaying();
	  
	  pthread_mutex_unlock (&_mutex);
	  return result;
	}
      

      virtual void setDevice(MidiDevice *dev)
	{
	  pthread_mutex_lock (&_mutex);
	  
	  MidiPlayer::setDevice(dev);
	  
	  pthread_mutex_unlock (&_mutex);
	}      

      virtual MidiDevice* getDevice()
	{
	  pthread_mutex_lock (&_mutex);
	  
	  MidiDevice* res = MidiPlayer::getDevice();
	  
	  pthread_mutex_unlock (&_mutex);
	  return res;
	}      

    protected:

      void eventAny(vector<pair<unsigned long, MidiEvent*> >::iterator i)   
	{
	  while (_manager->hasInternalEventWaiting())
	    {
	      if (_manager->internalRun() == false)
		return;
	    }

	  unsigned long ms = (getDevice()->ticksToMs(getFile()->getDivision(), i->first))/1000;
	  char buffer[8];
	  sprintf(buffer, "%02d:%02d", (int)(ms/60000), (int)((ms/1000)%60));
	  _manager->sendTick(i->first, buffer);
	}

      virtual void eventLyric(vector<pair<unsigned long, MidiEvent*> >::iterator i)
	{
	  _manager->sendLyrics(i->first);
	}

      virtual void eventText(vector<pair<unsigned long, MidiEvent*> >::iterator i)
	{
	  _manager->sendLyrics(i->first);
	}

      virtual void eventEnd(vector<pair<unsigned long, MidiEvent*> >::iterator i)
	{
	  _manager->sendEnd();
	}

    };

  friend class LocalMidiPlayer;
  friend void *midnmanager_run(void *data);

 public:

  enum {
    UNKOWN, TICK_MS, CLEAR_LYRICS_TEXT, ADD_LYRICS_TEXT, TICK_LYRICS, SET_FILE_LENGTH, END_LYRICS_TEXT, SONG_ENDED,
    CLEAR_TITLE_TEXT, ADD_TITLE_TEXT, END_TITLE_TEXT,
    CLEAR_INFO_TEXT, ADD_INFO_TEXT, END_INFO_TEXT,
    SET_LANGUAGE, UPDATE_INFO
  };

 public:

  MidnightPlayerManager(MidnightConfiguration*config);
  virtual ~MidnightPlayerManager()
    {
    }

  virtual void configurationChanged()
    {
      if (_player == NULL)
	return;

      _config->unlockDevice(_player->getDevice());
      
      MidiDevice *nudev = _config->lockDevice();
      _player->setDevice(nudev);
    }
  

  void saveConfig()
    {
      lockConfiguration();
      _config->save();
      unlockConfiguration();
    }

  //
  //
  //
  void setSelectedFile(string buffer)
    {
      _playerqueue.push(InfoBag(PLAYER_LOAD, 0, buffer));
    }

  void setPosition(int ticks)
    {
      _playerqueue.push(InfoBag(PLAYER_SET_POSITION, ticks));
    }

  int ticksToMs(int ticks)
    {
      unsigned long ms = (_device->ticksToMs(_file->getDivision(), ticks))/1000;
      return ms;
    }

  bool isPlaying()
    {
      if (_player != NULL)
	return _player->isPlaying();
      return false;
    }

  bool checkNotPlaying()
    {
      if (_player == 0)
	return true;

      if (_player->isPossiblyPlaying() == false)
	return true;
      else return false;
    }

  inline void stop()
    {
      _playerqueue.push(InfoBag(PLAYER_STOP, 0));
    }

  inline void play()
    {
      _playerqueue.push(InfoBag(PLAYER_PLAY, 0));
    }
  
  inline void pause()
    {
      _playerqueue.push(InfoBag(PLAYER_PAUSE, 0));
    }

  int pollEvent(int &value1, string &value2)
    {
      if (_controllerqueue.empty())
	return CONTROLLER_NOTHING;
      
      InfoBag bag = _controllerqueue.pop();

      switch(bag.Command)
	{
	case CONTROLLER_TICK:
	  {
	    value1 = bag.Value; 
	    value2 = bag.Buffer;
	  }
	  return TICK_MS;

	case CONTROLLER_CLEAR_LYRICS:
	  return CLEAR_LYRICS_TEXT;

	case CONTROLLER_ADD_LYRICS:
	  {
	    value1 = bag.Value;
	    value2 = bag.Buffer;
	  }
	  return ADD_LYRICS_TEXT;

	case CONTROLLER_ADD_LYRICS_END:
	  return END_LYRICS_TEXT;

	case CONTROLLER_CLEAR_TITLE:
	  return CLEAR_TITLE_TEXT;
	case CONTROLLER_ADD_TITLE:
	  {
	    value2 = bag.Buffer;
	  }
	  return ADD_TITLE_TEXT;
	case CONTROLLER_ADD_TITLE_END:
	  return END_TITLE_TEXT;


	case CONTROLLER_CLEAR_INFO:
	  return CLEAR_INFO_TEXT;
	case CONTROLLER_ADD_INFO:
	  {
	    value2 = bag.Buffer;
	  }
	  return ADD_INFO_TEXT;
	case CONTROLLER_ADD_INFO_END:
	  return END_INFO_TEXT;

	case CONTROLLER_SET_LANGUAGE:
	  {
	    value2 = bag.Buffer;
	  }
	  return SET_LANGUAGE;

	case CONTROLLER_TICK_LYRICS:
	  {
	    value1 = bag.Value;
	  }
	  return TICK_LYRICS;

	case CONTROLLER_TOTAL_LENGTH:
	  {
	    value1 = bag.Value;
	  }
	  return SET_FILE_LENGTH;

	case CONTROLLER_SONG_ENDED:
	  //	  cerr << "received song ended" << endl;
	  return SONG_ENDED;

	case CONTROLLER_UPDATE_INFO:
	  return UPDATE_INFO;
	}
      
      return UNKOWN;
    }

  void flush()
    {
      while (_playerqueue.empty() == false)
	usleep(1);
    }


 private:

  bool hasInternalEventWaiting()
    {
      return _playerqueue.empty() == false;
    }

      
  //
  // returns : false if important changes occurred (i.e. a new song
  // was loaded).
  //
  bool internalRun()
    {
      if (_playerqueue.empty())
	return true;

      while(_playerqueue.empty() == false)
	{
	  if (_playerqueue.empty() == true)
	    {
	      perror("ERROOR");
	    }

	  InfoBag ib = _playerqueue.pop();
	  //	InfoBag ib = _playerqueue.flavor();

	  switch(ib.Command)
	    {
	    case PLAYER_LOAD:
	      {
		//		cerr << "internalRun : load <" << ib.Buffer << ">" << endl;
		// load a new file
		if (_player != NULL)
		  _player->stop();
		load(ib.Buffer);

		//	      _playerqueue.pop();
		////////////////////////////////////////////////		return false;
	      }
	      break;
	  
	    case PLAYER_PLAY:
	      // play
	      if (_player != NULL)
		_player->play();

	      break;

	    case PLAYER_PAUSE:
	      // pause
	      if (_player != NULL)
		_player->pause();
	      break;

	    case PLAYER_STOP:
	      if (_player != NULL)
		_player->stop();
	      checkNotPlaying();
	      break;

	    case PLAYER_SET_POSITION:
	      // set position
	      {
		if (_player != NULL)
		  {
		    _player->setPosition(ib.Value);
		  }
	      }
	      break;
	    }
	}

      //_playerqueue.pop();

      return true;
    }


  void load(string filename);

  void sendFileInformation();

  void sendTick(unsigned long tick, const char *timestr)
    {
      _controllerqueue.push(InfoBag(CONTROLLER_TICK, tick, timestr));
    }

  void sendLyrics(unsigned long tick)
    {
      _controllerqueue.push(InfoBag(CONTROLLER_TICK_LYRICS, tick));
    }

  void sendEnd()
    {
      _controllerqueue.push(InfoBag(CONTROLLER_SONG_ENDED, 0));
    }


 protected:
  bool _configLock;

 public:
  
  MidnightConfiguration* lockConfiguration()
    {
      if (_configLock == true)
	return NULL;
      _configLock = true;
      return _config;
    }

  void unlockConfiguration()
    {
      _configLock = false;
    }
  
};

#endif
